// TODO :       

unit FlatMonthControl;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls;

type

  TButtonClickEvent = procedure (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer) of object;

  TLookAndFeelBehaviour = (lfStandard, lfFlat);

  TMonthSize = 1..12;
  TYearSize = 1..MaxInt;

  TMonthYearDate = class(TPersistent)
  private
    FMonth: TMonthSize;
    FYear: TYearSize;
    FIntDate: TYearSize;
    FOnChange: TNotifyEvent;
    FUpdateDate: Boolean;
    procedure SetMonth(const AValue: TMonthSize);
    procedure SetYear(const AValue: TYearSize);
    function GetIntDate: TYearSize;
    procedure SetUpdateDate(const Value: Boolean);
  protected
    procedure Change; dynamic;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property UpdateDate: Boolean read FUpdateDate write SetUpdateDate;
  public
    constructor Create;
    function ParseDateToInt(AMonth: TMonthSize; AYear: TYearSize): TYearSize;
    function ParseDateToStr(var ADate: TYearSize): String;
    procedure SetIntDate(ADate: TYearSize; var AMonth: TMonthSize; var AYear: TYearSize);
    property IntDate: TYearSize read GetIntDate;
  published
    property Month: TMonthSize read FMonth write SetMonth;
    property Year: TYearSize read FYear write SetYear;
  end;

  TCustomFlatMonthControl = class(TCustomControl)
  private
    FNewMonth: TMonthSize;
    FNewYear: TYearSize;
    FMaxLongMonthName: String;
    FDate: TMonthYearDate;
    FUseSystemFont: Boolean;
    FNeedRedrawButton: Boolean;
    FRepeatPause: Integer;
    FRepeatTimer: TTimer;
    FHideSelection: Boolean;
    FLookAndFeel: TLookAndFeelBehaviour;
    FTopDate: TYearSize;
    FItemIndex: Integer;
    FUpdatingDate: Boolean;
    FTopButtonPressed: Boolean;
    FBottomButtonPressed: Boolean;
    FOnTopButtonPressed: Boolean;
    FOnBottomButtonPressed: Boolean;
    FRowHeight: Integer;
    FGridColor: TColor;
    FSelectedColor: TColor;
    FHighLightTextColor: TColor; //clHighLightText
    FOnChange: TNotifyEvent;
    FUpdateDate: Boolean;
    FOnTopButtonMouseDown: TButtonClickEvent;
    FOnBottomButtonMouseDown: TButtonClickEvent;
    FOnTopButtonMouseUp: TButtonClickEvent;
    FOnBottomButtonMouseUp: TButtonClickEvent;
    function GetDateStr: String;
    function GetMaxLines: Integer;
    function GetTopButtonRect: TRect;
    function GetBottomButtonRect: TRect;
    function GetCentralizeItemIndex: Integer;
    function GetSelectedIntDateRect(X, Y: Integer): Integer;
    function GetItemIndexFromCoord(X, Y: Integer): Integer;
    //function SetSizeRowHeight: Integer;
    procedure SetSelectionToCenter;
    procedure SetUseSystemFont(const AValue: Boolean);
    procedure SetLookAndFeel(const AValue: TLookAndFeelBehaviour);
    procedure RefreshDateInformation;
    procedure TimerExpired(Sender: TObject);
    procedure SetRowHeight(const AValue: Integer);
    procedure SetHideSelection(const AValue: Boolean);
    procedure RepaintTopButton;
    procedure RepaintBottomButton;
    procedure DrawButton(ATopButton, APressed: Boolean; FLookAndFeel: TLookAndFeelBehaviour);
    procedure CMWinIniChange(var Message: TWMWinIniChange); message CM_WININICHANGE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
    procedure CMEnter(var Msg: TCMEnter); message CM_ENTER;
    procedure CMExit(var Msg: TCMExit); message CM_EXIT;
    procedure WMKillFocus(var Msg: TWMSetFocus); message WM_KILLFOCUS;
    procedure WMMouseActivate(var Msg: TWMMouseActivate); message WM_MOUSEACTIVATE;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure ResizeWindow;
    procedure SetGridColor(const AValue: TColor);
    procedure SetHighLightTextColor(const AValue: TColor);
    procedure SetSelectedColor(const AValue: TColor);
    //procedure Up;
    //procedure Down;
    procedure MoveBy(Distance: Integer);
    function GetFirstDay: TDateTime;
    function GetLastDay: TDateTime;
    procedure SetUpdateDate(const Value: Boolean);
    function GetUpdateDate: Boolean;
    procedure TopButtonMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure BottomButtonMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure TopButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure BottomButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  protected
    procedure Change; dynamic;
    procedure CreateWnd; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure SetTopDate(Sender: TObject);
    procedure SetItemIndex;
    procedure DrawDateDirectMode(R: TRect; S: String; TextColor: TColor);
    property OnTopButtonMouseDown: TButtonClickEvent read FOnTopButtonMouseDown write FOnTopButtonMouseDown;
    property OnBottomButtonMouseDown: TButtonClickEvent read FOnBottomButtonMouseDown write FOnBottomButtonMouseDown;
    property OnTopButtonMouseUp: TButtonClickEvent read FOnTopButtonMouseUp write FOnTopButtonMouseUp;
    property OnBottomButtonMouseUp: TButtonClickEvent read FOnBottomButtonMouseUp write FOnBottomButtonMouseUp;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // Standard
    property Align;
    property Color;
    property Font;
    property TabOrder;
    property TabStop;
    property Visible;
    // Extra
    property GridColor: TColor read FGridColor write SetGridColor;
    property SelectedColor: TColor read FSelectedColor write SetSelectedColor;
    property HighLightTextColor: TColor read FHighLightTextColor write SetHighLightTextColor; //clHighLightText
    property UseSystemFont: Boolean read FUseSystemFont write SetUseSystemFont;
    property LookAndFeel: TLookAndFeelBehaviour read FLookAndFeel write SetLookAndFeel;
    property HideSelection: Boolean read FHideSelection write SetHideSelection;
    property RowHeight: Integer read FRowHeight write SetRowHeight;
    property RepeatPause: Integer read FRepeatPause write FRepeatPause;  { pause before repeat timer (ms) }
    property MaxLines: Integer read GetMaxLines;
    property ItemIndex: Integer read FItemIndex;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    // Date
    property Date: TMonthYearDate read FDate write FDate;
    property DateStr: String read GetDateStr;
    property FirstDay: TDateTime read GetFirstDay;
    property LastDay: TDateTime read GetLastDay;
    property UpdateDate: Boolean read GetUpdateDate write SetUpdateDate;
  end;

  TFlatMonthControl = class(TCustomFlatMonthControl)
  public
    property ItemIndex;
  published
    // Standard
    property Align;
    property Color;
    property Font;
    property TabOrder;
    property TabStop default True;
    property Visible;
    // Extra
    property UseSystemFont;
    property GridColor;
    property SelectedColor;
    property HighLightTextColor;
    property LookAndFeel;
    property HideSelection default False;
    property RowHeight;
    property RepeatPause;
    property OnChange;
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnKeyPress;
    property OnKeyDown;
    property OnKeyUp;
    property OnTopButtonMouseDown;
    property OnBottomButtonMouseDown;
    property OnTopButtonMouseUp;
    property OnBottomButtonMouseUp;
    // Date
    property Date;
  end;

implementation

{ TCustomFlatMonthControl }

constructor TCustomFlatMonthControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FUpdateDate := True;
  FDate := TMonthYearDate.Create;
  FDate.OnChange := SetTopDate;
  FMaxLongMonthName := '';
  FGridColor := clSilver;
  FSelectedColor := clActiveCaption;
  FHighLightTextColor := clHighLightText;
  FNewMonth := 1;
  FNewYear := 1;
  Width := 130;
  Height := 178;
  FRowHeight := 16;
  FRepeatPause := 100;
  BevelInner := bvNone;
  BevelOuter := bvLowered;
  BevelKind := bkTile;
  FLookAndFeel := lfFlat;
  ParentColor := False;
  Color := clWindow;
  TabStop := True;
  ControlStyle := ControlStyle + [csDesignInteractive];
end;

procedure TCustomFlatMonthControl.CreateWnd;
begin
  inherited CreateWnd;
  FItemIndex := GetCentralizeItemIndex;
  FUpdatingDate := True;
  SetTopDate(nil);
  FUpdatingDate := False;
end;

destructor TCustomFlatMonthControl.Destroy;
begin
  if FRepeatTimer <> nil then
    FRepeatTimer.Free;
  if FDate <> nil then
    FDate.Free;
  inherited Destroy;
end;

procedure TCustomFlatMonthControl.RefreshDateInformation;
var
  DefaultLCID: LCID;
  I, Day: Integer;
begin
  GetFormatSettings;
  DefaultLCID := GetUserDefaultLCID;
  for I := 1 to 12 do
  begin
    ShortMonthNames[I] :=
      GetLocaleStr(DefaultLCID, LOCALE_SABBREVMONTHNAME1 + I - 1, ShortMonthNames[I]);
    LongMonthNames[I] :=
      GetLocaleStr(DefaultLCID, LOCALE_SMONTHNAME1 + I - 1, LongMonthNames[I]);
    if Length(LongMonthNames[I]) > Length(FMaxLongMonthName) then
      FMaxLongMonthName := LongMonthNames[I];
  end;
  for I := 1 to 7 do
  begin
    Day := I - 2;
    if Day = -1 then Day := 6;
    ShortDayNames[I] :=
      GetLocaleStr(DefaultLCID, LOCALE_SABBREVDAYNAME1 + Day, ShortDayNames[I]);
    LongDayNames[I] :=
      GetLocaleStr(DefaultLCID, LOCALE_SDAYNAME1 + Day, LongDayNames[I]);
  end;
  TimeAMString := GetLocaleStr(DefaultLCID, LOCALE_S1159, TimeAMString);
  TimePMString := GetLocaleStr(DefaultLCID, LOCALE_S2359, TimePMString);
end;


procedure TCustomFlatMonthControl.CMWinIniChange(var Message: TWMWinIniChange);
begin
  inherited;
  RefreshDateInformation;
  Invalidate;
end;

procedure TCustomFlatMonthControl.TimerExpired(Sender: TObject);
var
  MousePos: TPoint;
begin
  FRepeatTimer.Interval := FRepeatPause;
  try
    MousePos := Point(Mouse.CursorPos.X, Mouse.CursorPos.Y);
    MouseDown(mbLeft, [], ScreenToClient(MousePos).X, ScreenToClient(MousePos).Y);
  except
    FRepeatTimer.Enabled := False;
    FNeedRedrawButton := True;
    raise;
  end;
end;

function TCustomFlatMonthControl.GetBottomButtonRect: TRect;
begin
  SetRect(Result, 0, ClientHeight - 16, ClientWidth, ClientHeight);
end;

function TCustomFlatMonthControl.GetTopButtonRect: TRect;
begin
  SetRect(Result, 0, 0, ClientWidth, 16);
end;

function TCustomFlatMonthControl.GetSelectedIntDateRect(X, Y: Integer): Integer;
begin
  Result := FTopDate + ((Y - 16) div FRowHeight);
end;

function TCustomFlatMonthControl.GetItemIndexFromCoord(X, Y: Integer): Integer;
begin
  Result := ((Y - 16) div FRowHeight);
{  if (Result = MaxLines) then
    Result := FItemIndex;}
end;

procedure TCustomFlatMonthControl.SetRowHeight(const AValue: Integer);
begin
 if (FRowHeight = AValue) then Exit;
 if (AValue < Font.Size + 4) then
   FRowHeight := Font.Size + 4
 else
   FRowHeight := AValue;  
{  if (FRowHeight <> AValue) then
  begin
    if (SetSizeRowHeight > AValue) then
      FRowHeight := SetSizeRowHeight
    else
      FRowHeight := AValue;
    Invalidate;
  end;}
end;

procedure TCustomFlatMonthControl.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then
  begin
    if PtInRect(GetTopButtonRect, Point(X, Y)) then
    begin
      begin
      FNeedRedrawButton := False;
        if FRepeatTimer = nil then
          FRepeatTimer := TTimer.Create(Self);

        FRepeatTimer.OnTimer := TimerExpired;
        FRepeatTimer.Interval := FRepeatPause;
        FRepeatTimer.Enabled  := True;
      end;
      FOnTopButtonPressed := True;
      FOnBottomButtonPressed := False;
      FTopButtonPressed := True;
      if FNeedRedrawButton then RepaintTopButton;
      TopButtonMouseDown(Self, Button, Shift, X, Y);
      MoveBy(-1);
      //Up;
      Exit;
    end;
    if PtInRect(GetBottomButtonRect, Point(X, Y)) then
    begin
      begin
      FNeedRedrawButton := False;
        if FRepeatTimer = nil then
          FRepeatTimer := TTimer.Create(Self);

        FRepeatTimer.OnTimer := TimerExpired;
        FRepeatTimer.Interval := FRepeatPause;
        FRepeatTimer.Enabled  := True;
      end;
      FOnTopButtonPressed := False;
      FOnBottomButtonPressed := True;
      FBottomButtonPressed := True;
      if FNeedRedrawButton then RepaintBottomButton;
      BottomButtonMouseDown(Self, Button, Shift, X, Y);
      MoveBy(1);
      //Down;
      Exit;
    end;

    if (GetItemIndexFromCoord(X, Y) < MaxLines) then
    begin
      FItemIndex := GetItemIndexFromCoord(X, Y);
      FDate.SetIntDate(GetSelectedIntDateRect(X, Y), FNewMonth, FNewYear);
    end
    else
    begin
      FNewMonth := FDate.Month;
      FNewYear := FDate.Year;
    end;

    FOnTopButtonPressed := False;
    FOnBottomButtonPressed := False;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TCustomFlatMonthControl.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if (ssLeft in Shift) then
  begin

    if (FTopButtonPressed <> PtInRect(GetTopButtonRect, Point(X, Y))) and FOnTopButtonPressed then
      begin
        if FRepeatTimer <> nil then
          FRepeatTimer.Enabled  := False;
        FTopButtonPressed := not FTopButtonPressed;
        if FTopButtonPressed then MouseDown(mbLeft, Shift, X, Y);
        RepaintTopButton;
        Exit;
      end;

    if (FBottomButtonPressed <> PtInRect(GetBottomButtonRect, Point(X, Y))) and FOnBottomButtonPressed then
      begin
        if FRepeatTimer <> nil then
          FRepeatTimer.Enabled  := False;
        FBottomButtonPressed := not FBottomButtonPressed;
        if FBottomButtonPressed then MouseDown(mbLeft, Shift, X, Y);
        RepaintBottomButton;
        Exit;
      end;
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TCustomFlatMonthControl.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
if FRepeatTimer <> nil then
    FRepeatTimer.Enabled  := False;
  if Button = mbLeft then
  begin
    if FTopButtonPressed then
    begin
      FTopButtonPressed := False;
      RepaintTopButton;
      TopButtonMouseUp(Self, Button, Shift, X, Y);
      Exit;
    end;
    if FBottomButtonPressed then
    begin
      FBottomButtonPressed := False;
      RepaintBottomButton;
      BottomButtonMouseUp(Self, Button, Shift, X, Y);
      Exit;
    end;
    FDate.Month := FNewMonth;
    FDate.Year := FNewYear;
    SetSelectionToCenter;
  end;
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TCustomFlatMonthControl.RepaintTopButton;
var
  R: TRect;
begin
  R := GetTopButtonRect;
  InvalidateRect(Handle, @R, False);
end;

procedure TCustomFlatMonthControl.RepaintBottomButton;
var
  R: TRect;
begin
  R := GetBottomButtonRect;
  InvalidateRect(Handle, @R, False);
end;

procedure TCustomFlatMonthControl.DrawButton(ATopButton, APressed: Boolean; FLookAndFeel: TLookAndFeelBehaviour);
const
  BordersLook: array[TLookAndFeelBehaviour] of Boolean = (False, True);
var
  ClipRgnExists: Boolean;
  R: TRect;
  DC: HDC;
  Rgn1, Rgn2: HRGN;
  Kind: UINT;
begin
  if ATopButton then
    R := GetTopButtonRect
  else
    R := GetBottomButtonRect;
  DC := Canvas.Handle;
  Rgn1 := 0;
  ClipRgnExists := False;

  if BordersLook[FLookAndFeel] then
  begin
    Rgn1 := CreateRectRgn(0, 0, 0, 0);
    ClipRgnExists := GetClipRgn(DC, Rgn1) = 1;
    InflateRect(R, -1, -1);
    Rgn2 := CreateRectRgnIndirect(R);
    SelectClipRgn(DC, Rgn2);
    DeleteObject(Rgn2);
    InflateRect(R, 1, 1);
  end;

  if ATopButton then
    Kind := DFCS_SCROLLUP
  else
    Kind := DFCS_SCROLLDOWN;

  if BordersLook[FLookAndFeel] then
  begin
    DrawFrameControl(DC, R, DFC_SCROLL,
      Kind or Byte(APressed) * DFCS_PUSHED or Byte(BordersLook[FLookAndFeel]) * DFCS_FLAT);
    if ClipRgnExists then
      SelectClipRgn(DC, Rgn1)
    else
      SelectClipRgn(DC, 0);
    DeleteObject(Rgn1);
    DrawEdge(DC, R, BDR_RAISEDINNER, BF_RECT);
    if APressed then InvertRect(DC, R);
  end
  else
    DrawFrameControl(DC, R, DFC_SCROLL,
      Kind or Byte(APressed) * DFCS_PUSHED or Byte(BordersLook[FLookAndFeel]) * DFCS_FLAT);

end;

procedure TCustomFlatMonthControl.Paint;

  procedure DrawLine(X1, Y1, X2, Y2: integer);
  begin
    Canvas.MoveTo(X1, Y1);
    Canvas.LineTo(X2, Y2);
  end;

var
  I: Integer;
  R: TRect;
  S: String;
  OldBkColor, OldTextColor: Integer;
  TextColor: TColor;
  IntTopDate: TYearSize;
  Rem: Integer;
begin
  Canvas.Pen.Color := FGridColor;
  Canvas.Font := Font;
//  FRowHeight := SetSizeRowHeight;
  IntTopDate := FTopDate;
  for I := 0 to MaxLines -1 do
  begin
    if I = FItemIndex then
    begin
      if not FHideSelection then
      begin
        Canvas.Brush.Color := clSilver;
      end;
      if Focused then
        Canvas.Brush.Color := FSelectedColor;
    end
    else
    begin
      Canvas.Brush.Color := Color;
    end;

    R := Rect(0, I*FRowHeight + 16, ClientWidth, (I+1)*FRowHeight + 16);

    S := FDate.ParseDateToStr(IntTopDate);

    TextColor := Font.Color;

    if (Focused) and (I = FItemIndex) then
      TextColor := FHighLightTextColor;
    DrawDateDirectMode(R, S, TextColor);
    if IntTopDate <> MaxInt then
      Inc(IntTopDate);

    if (Focused) and (I = FItemIndex) then
    begin
      OldBkColor := SetBkColor(Canvas.Handle, clBlack);
      OldTextColor := SetTextColor(Canvas.Handle, clWhite);
      Canvas.DrawFocusRect(R);
      SetBkColor(Canvas.Handle, OldBkColor);
      SetTextColor(Canvas.Handle, OldTextColor);
    end
    else
    begin
      DrawLine(R.Left, R.Bottom - 1, R.Right, R.Bottom - 1);
      DrawLine(R.Right - 1, R.Bottom - 1, R.Right - 1, R.Top - 1);
    end;

  end;
    DrawButton(True, FTopButtonPressed, FLookAndFeel);
    DrawButton(False, FBottomButtonPressed, FLookAndFeel);

  Rem := ClientHeight - 32 - (FRowHeight)*MaxLines;
  if (MaxLines > 0) then
  begin
    Canvas.Brush.Color := Color;
    if (Rem <> 0) then
      Canvas.FillRect(Rect(0, ClientHeight - 16 - Rem, ClientWidth, ClientHeight - 16));
  end;
end;

procedure TCustomFlatMonthControl.DrawDateDirectMode(R: TRect; S: String; TextColor: TColor);
begin
  Canvas.Font.Color := TextColor;
  FillRect(Canvas.Handle, R, Canvas.Brush.Handle);
  SetBkMode(Canvas.Handle, TRANSPARENT);
//  ExtTextOut(Canvas.Handle, (R.Right - Canvas.TextWidth(S)) div 2, R.Top + (FRowHeight - Font.Size) div 2 - 3, ETO_CLIPPED, @R, PChar(S), Length(S), nil);
  DrawText(Canvas.Handle, PChar(S), -1, R, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
//  DrawCellText(Canvas.Handle, R, Canvas.Brush.Handle, TextColor, nil, S, Length(S), False, False, taCenter, False, False, False);
end;

(*function TCustomFlatMonthControl.SetSizeRowHeight: Integer;
var
  TextMetric: TTextMetric;
begin
//  Result := 1{FRowHeight};
  GetTextMetrics(Canvas.Handle, TextMetric);

  if (FRowHeight{Result} < TextMetric.tmHeight + TextMetric.tmInternalLeading) then
    Result := TextMetric.tmHeight + TextMetric.tmInternalLeading
  else
    Result := FRowHeight;
end;*)


procedure TCustomFlatMonthControl.SetHideSelection(const AValue: Boolean);
begin
  if FHideSelection <> AValue then
  begin
    FHideSelection := AValue;
    Invalidate;
  end;
end;

procedure TCustomFlatMonthControl.WMEraseBkgnd(var Message: TWmEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TCustomFlatMonthControl.CMEnter(var Msg: TCMEnter);
begin
  inherited;
  Invalidate;
end;

procedure TCustomFlatMonthControl.CMExit(var Msg: TCMExit);
begin
  inherited;
  Invalidate;
end;

procedure TCustomFlatMonthControl.WMKillFocus(var Msg: TWMSetFocus);
begin
  inherited;
  Invalidate;
end;

procedure TCustomFlatMonthControl.WMMouseActivate(var Msg: TWMMouseActivate);
begin
  inherited;
  Windows.SetFocus(Handle);
  Invalidate;
end;

procedure TCustomFlatMonthControl.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  Invalidate;
end;

// Setting FItemIndex

function TCustomFlatMonthControl.GetMaxLines: Integer;
begin
  Result := 0;
  Inc(Result, (ClientHeight - 32) div (FRowHeight));
end;

function TCustomFlatMonthControl.GetCentralizeItemIndex: Integer;
begin
  Result := MaxLines div 2;
end;

procedure TCustomFlatMonthControl.SetSelectionToCenter;
begin
  FUpdatingDate := True;
  SetTopDate(nil);
  FUpdatingDate := False;
end;

// Color, Looks, Sizing

procedure TCustomFlatMonthControl.SetLookAndFeel(const AValue: TLookAndFeelBehaviour);
begin
  if AValue <> FLookAndFeel then
  begin
    FLookAndFeel := AValue;
    if AValue = lfStandard then
    begin
      BevelInner := bvLowered;
      BevelOuter := bvLowered;
      BevelKind := bkTile;
    end
    else
    begin
      BevelInner := bvNone;
      BevelOuter := bvLowered;
      BevelKind := bkTile;
    end;
    RecreateWnd;
  end;
end;

procedure TCustomFlatMonthControl.WMSize(var Message: TWMSize);
begin
  inherited;
  ResizeWindow;
end;

procedure TCustomFlatMonthControl.ResizeWindow;
var
//  Size: TSize;
  S: String;
  TextMetric: TTextMetricA;
begin
  RefreshDateInformation;
  S := FMaxLongMonthName + ' ' + IntToStr(FDate.Month);
  if ClientHeight < 32 + FRowHeight then
    ClientHeight := 32 + FRowHeight;
//  GetTextExtentPoint32(Canvas.Handle, PChar(S), Length(S), Size);

  GetTextMetrics(Canvas.Handle, TextMetric);
{  if ClientWidth < (TextMetric.tmMaxCharWidth)*Length(S) + 8 then
    ClientWidth := (TextMetric.tmMaxCharWidth)*Length(S) + 8;}
  SetSelectionToCenter;
end;

procedure TCustomFlatMonthControl.CMFontChanged(var Message: TMessage);
begin
//  FRowHeight := SetSizeRowHeight;
  SetSelectionToCenter;
  Invalidate;
//  ShowMessage(IntToStr(Font.Size));
end;

procedure TCustomFlatMonthControl.SetUseSystemFont(const AValue: Boolean);
var
  NonClientMetrics: TNonClientMetrics;
begin
  if FUseSystemFont <> AValue then
  begin
    FUseSystemFont := AValue;
    if FUseSystemFont then
    begin
      NonClientMetrics.cbSize := SizeOf(NonClientMetrics);
      SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0);
      if FUseSystemFont then
      begin
        Font.Handle := CreateFontIndirect(NonClientMetrics.lfMenuFont);
        FUseSystemFont := True;
      end;
    end;
  end;
end;

procedure TCustomFlatMonthControl.SetItemIndex;
begin
  if (MaxLines = 1) and (MaxInt = FDate.IntDate) then
  begin
    FItemIndex := 0;
    Exit;
  end;

//  if (FItemIndex <= (MaxLines div 2)) then
  if (FTopDate = 1) then
  begin
    FItemIndex := FDate.IntDate - FTopDate;
  end else
  if ((MaxInt - FDate.IntDate) < (MaxLines div 2)) then
  begin
    if (MaxInt <> FDate.IntDate) then
    begin
      FItemIndex := MaxLines - MaxInt + FDate.IntDate - 1
    end
    else
    begin
      FItemIndex := MaxLines - 1;
    end;
  end else
  if (MaxInt <> FDate.IntDate) then
  begin
    FItemIndex := (MaxLines div 2);
  end;
end;

procedure TCustomFlatMonthControl.SetTopDate;
var
  I: Integer;
begin
  I := FDate.IntDate - (MaxLines div 2);
  if (I < 1) then
  begin
    FTopDate := 1;
  end else
  if (I > (MaxInt - MaxLines + 1)) then
  begin
      FTopDate := (MaxInt - MaxLines + 1);
  end else
  begin
    FTopDate := I;
  end;
  SetItemIndex;
  Invalidate;
  //if not (csDesigning in ComponentState) or not (csLoading in ComponentState)then
  if (not FUpdatingDate) then
  begin
    if (not FUpdateDate) then
    begin
      Change;
    end;
  end;
end;

procedure TCustomFlatMonthControl.SetGridColor(const AValue: TColor);
begin
  if FGridColor <> AValue then
  begin
    FGridColor := AValue;
    Invalidate;
  end;
end;

procedure TCustomFlatMonthControl.SetHighLightTextColor(const AValue: TColor);
begin
  if FHighLightTextColor <> AValue then
  begin
    FHighLightTextColor := AValue;
    Invalidate;
  end;
end;

procedure TCustomFlatMonthControl.SetSelectedColor(const AValue: TColor);
begin
  if FSelectedColor <> AValue then
  begin
    FSelectedColor := AValue;
    Invalidate;
  end;
end;

procedure TCustomFlatMonthControl.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TCustomFlatMonthControl.KeyDown(var Key: Word; Shift: TShiftState);
begin
  { TODO : FRepeatTimer }
  case Key of
    VK_Left, VK_Up    : MoveBy(-1);//Up;
    VK_Right, VK_Down : MoveBy(1);//Down;
    VK_Prior:
      if (ssCtrl in Shift) then
          MoveBy(-MaxLines)
       else
          MoveBy(-12);
    Vk_Next:
       if (ssCtrl in Shift) then
          MoveBy(MaxLines)
       else
          MoveBy(12);
    VK_Home:
      if (ssCtrl in Shift) then
      begin
        FDate.SetIntDate(1, FNewMonth, FNewYear);
        FDate.Month := FNewMonth;
        FDate.Year := FNewYear;
      end
      else
      begin
        FDate.SetIntDate(FTopDate, FNewMonth, FNewYear);
        FDate.Month := FNewMonth;
        FDate.Year := FNewYear;
      end;
    VK_End:
      if (ssCtrl in Shift) then
      begin
        FDate.SetIntDate(MaxInt, FNewMonth, FNewYear);
        FDate.Month := FNewMonth;
        FDate.Year := FNewYear;
      end
      else
      begin
        MoveBy(MaxLines - ItemIndex - 1);
      end;
  end;
  inherited;
end;

procedure TCustomFlatMonthControl.MoveBy(Distance: Integer);
var
  RealDistance: Integer;
  Difference: Integer;
begin
  if (Distance = 0) then Exit;

  Difference := (MaxInt - FDate.IntDate);
  
  if (Difference >= Distance) then
  begin
    RealDistance := FDate.IntDate + Distance;
  end
  else
  begin
    RealDistance := MaxInt;
  end;

  if (RealDistance < 1) then
    RealDistance := 1;

  FDate.SetIntDate(RealDistance, FNewMonth, FNewYear);
  FDate.Month := FNewMonth;
  FDate.Year := FNewYear;
end;


{procedure TCustomFlatMonthControl.Up;
begin
  if (FDate.IntDate > 1) then
  begin
    FDate.SetIntDate(FDate.IntDate - 1, FNewMonth, FNewYear);
    FDate.Month := FNewMonth;
    FDate.Year := FNewYear;
  end;
end;

procedure TCustomFlatMonthControl.Down;
begin
  if (FDate.IntDate < MaxInt) then
  begin
    FDate.SetIntDate(FDate.IntDate + 1, FNewMonth, FNewYear);
    FDate.Month := FNewMonth;
    FDate.Year := FNewYear;
  end;
end;}

procedure TCustomFlatMonthControl.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;  
end;

function TCustomFlatMonthControl.GetDateStr: String;
begin
  Result := LongMonthNames[Date.Month] + ' ' + IntToStr(Date.Year);
end;

function TCustomFlatMonthControl.GetFirstDay: TDateTime;
var
  Year, Month, Day: Word;
begin
  Year := Word(Date.Year);
  Month := Word(Date.Month);
  Day := 1;
  Result := EncodeDate(Year, Month, Day);
end;

function TCustomFlatMonthControl.GetLastDay: TDateTime;

  function IsLeapYear(AYear: Integer): Boolean;
  begin
    Result := (AYear mod 4 = 0) and ((AYear mod 100 <> 0) or (AYear mod 400 = 0));
  end;

  function DaysPerMonth(AYear, AMonth: Integer): Integer;
  const
    DaysInMonth: array[1..12] of Integer =
      (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
  begin
    Result := DaysInMonth[AMonth];
    if (AMonth = 2) and IsLeapYear(AYear) then Inc(Result);
  end;

begin
  Result := EncodeDate(Date.Year, Date.Month, DaysPerMonth(Date.Year, Date.Month));
end;

procedure TCustomFlatMonthControl.SetUpdateDate(const Value: Boolean);
begin
  if (FUpdateDate <> Value) then
  begin
    FDate.UpdateDate := Value;
    FUpdateDate := Value;
    if FUpdateDate then
    begin
      Change;
    end;
  end;
end;

function TCustomFlatMonthControl.GetUpdateDate: Boolean;
begin
  Result := FDate.UpdateDate;
  FUpdateDate := Result;
end;

procedure TCustomFlatMonthControl.TopButtonMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnTopButtonMouseDown) then FOnTopButtonMouseDown(Sender, Button, Shift, X, Y);
end;

procedure TCustomFlatMonthControl.BottomButtonMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnBottomButtonMouseDown) then FOnBottomButtonMouseDown(Sender, Button, Shift, X, Y);
end;

procedure TCustomFlatMonthControl.BottomButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnBottomButtonMouseUp) then FOnBottomButtonMouseUp(Sender, Button, Shift, X, Y);
end;

procedure TCustomFlatMonthControl.TopButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnTopButtonMouseUp) then FOnTopButtonMouseUp(Sender, Button, Shift, X, Y);
end;

{ TMonthYearDate }

procedure TMonthYearDate.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

constructor TMonthYearDate.Create;
var
  Year, Month, Day: Word;
begin
  inherited;
  FUpdateDate := True;
  DecodeDate(Now, Year, Month, Day);
  FYear := Year;
  FMonth := Month;
end;

function TMonthYearDate.GetIntDate: TYearSize;
begin
  Result := ParseDateToInt(FMonth, FYear);
end;

function TMonthYearDate.ParseDateToInt(AMonth: TMonthSize; AYear: TYearSize): TYearSize;
begin
  Result := (AYear - 1) * 12 + AMonth;
  FIntDate := Result;
end;

function TMonthYearDate.ParseDateToStr(var ADate: TYearSize): String;
var
  AMonth: Integer;
  AYear: Integer;
begin
   AYear := ADate div 12 + 1;
   AMonth := ADate mod 12;
   if AMonth = 0 then
   begin
     if ADate >= 12 then AYear := AYear - 1;
     AMonth := 12;
   end;
   Result := LongMonthNames[AMonth] + ' ' + IntToStr(AYear);
end;

procedure TMonthYearDate.SetIntDate(ADate: TYearSize; var AMonth: TMonthSize; var AYear: TYearSize);
var
  IntMonth: Integer;
  IntYear: Integer;
begin
  IntYear := ADate div 12 + 1;
  IntMonth := ADate mod 12;
  if IntMonth = 0 then
  begin
    if ADate >= 12 then IntYear := IntYear - 1;
    IntMonth := 12;
  end;
  AYear := IntYear;
  AMonth := IntMonth;
end;

procedure TMonthYearDate.SetMonth(const AValue: TMonthSize);
begin
  if (FMonth <> AValue) then
  begin
    FMonth := AValue;
    if (not FUpdateDate) then
    begin
      Change;
    end;
  end;
end;

procedure TMonthYearDate.SetYear(const AValue: TYearSize);
begin
  if (FYear <> AValue) then
  begin
    FYear := AValue;
    if (not FUpdateDate) then
    begin
      Change;
    end;
  end;
end;

procedure TMonthYearDate.SetUpdateDate(const Value: Boolean);
begin
  if (FUpdateDate <> Value) then
  begin
    FUpdateDate := Value;
  end;
end;

end.
