unit FlatPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, FlatUtils, FlatBitBtn;

type
  TFlatBorderStyle = (fbsFlat, fbsFrame);

  TFlatBorders = class(TPersistent)
  private
    FLeftBorder: Boolean;
    FTopBorder: Boolean;
    FRightBorder: Boolean;
    FBottomBorder: Boolean;
    FChangeBorder: TNotifyEvent;
    procedure SetBottomBorder(const Value: Boolean);
    procedure SetLeftBorder(const Value: Boolean);
    procedure SetRightBorder(const Value: Boolean);
    procedure SetTopBorder(const Value: Boolean);
    procedure ChangeBorder; dynamic;
  public
    constructor Create;
  published
    property LeftBorder: Boolean read FLeftBorder write SetLeftBorder;
    property TopBorder: Boolean read FTopBorder write SetTopBorder;
    property RightBorder: Boolean read FRightBorder write SetRightBorder;
    property BottomBorder: Boolean read FBottomBorder write SetBottomBorder;
    property OnChangeBorder: TNotifyEvent read FChangeBorder write FChangeBorder;
  end;

  TFlatPanel = class(TCustomControl)
  private
    FCloseButton: TFlatBitBtn;
    FBorders: TFlatBorders;
    FBorderStyle: TFlatBorderStyle;
    FShowSeparator: Boolean;
    FSeparatorHeight: Integer;
    FExtraFlatCloseButton: Boolean;
    FShowCloseButton: Boolean;
    FEnabledCloseButton: Boolean;
    FOnCloseButtonClick: TNotifyEvent;
    //FButtonWidth: Integer;
    //FButtonHeight: Integer;
    FRightButtonOffset: Integer;
    FTopButtonOffset: Integer;
    function RectButton: TRect;
//    function MouseOnButton: Boolean;
    procedure SetExtraFlatCloseButton(Value: Boolean);
    procedure SetShowCloseButton(Value: Boolean);
    procedure SetBorderStyle(const Value: TFlatBorderStyle);
    procedure SetBorders(const Value: TFlatBorders);
    procedure SetShowSeparator(const Value: Boolean);
    procedure SetSeparatorHeight(const Value: Integer);
    procedure SetEnabledCloseButton(Value: Boolean);
    //procedure DrawCloseButton(APressed: Boolean);
    procedure SetRightButtonOffset(Value: Integer);
    procedure SetTopButtonOffset(Value: Integer);
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure SetCloseButtonTabStop(const Value: Boolean);
    function GetCloseButtonTabStop: Boolean;
    //procedure DrawCloseBitmap;
  protected
    procedure RepaintBorders(Sender: TObject);
    procedure DoCloseButtonClick(Sender: TObject); dynamic;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Align;
    property Anchors;
    property Borders: TFlatBorders read FBorders write SetBorders;
    property BorderStyle: TFlatBorderStyle read FBorderStyle write SetBorderStyle;
    property ShowSeparator: Boolean read FShowSeparator write SetShowSeparator;
    property Ctl3D;
    property Color;
    property ExtraFlatCloseButton: Boolean read FExtraFlatCloseButton write SetExtraFlatCloseButton;
    property Font;
    property ParentColor;
    property SeparatorHeight: Integer read FSeparatorHeight write SetSeparatorHeight;
    property ShowCloseButton: Boolean read FShowCloseButton write SetShowCloseButton;
    property EnabledCloseButton: Boolean read FEnabledCloseButton write SetEnabledCloseButton;
    property OnCloseButtonClick: TNotifyEvent read FOnCloseButtonClick write FOnCloseButtonClick;
    property RightButtonOffset: Integer read FRightButtonOffset write SetRightButtonOffset;
    property TopButtonOffset: Integer read FTopButtonOffset write SetTopButtonOffset;
    property CloseButtonTabStop: Boolean read GetCloseButtonTabStop write SetCloseButtonTabStop;
    property TabOrder;
    property TabStop;
  end;

implementation

{ TFlatBorders }

constructor TFlatBorders.Create;
begin
  FLeftBorder := True;
  FTopBorder := True;
  FRightBorder := True;
  FBottomBorder := True;
end;

procedure TFlatBorders.SetBottomBorder(const Value: Boolean);
begin
  if (FBottomBorder <> Value) then
  begin
    FBottomBorder := Value;
    ChangeBorder;
  end;
end;

procedure TFlatBorders.SetLeftBorder(const Value: Boolean);
begin
  if (FLeftBorder <> Value) then
  begin
    FLeftBorder := Value;
    ChangeBorder;
  end;
end;

procedure TFlatBorders.SetRightBorder(const Value: Boolean);
begin
  if (FRightBorder <> Value) then
  begin
    FRightBorder := Value;
    ChangeBorder;
  end;
end;

procedure TFlatBorders.SetTopBorder(const Value: Boolean);
begin
  if (FTopBorder <> Value) then
  begin
    FTopBorder := Value;
    ChangeBorder;
  end;
end;

procedure TFlatBorders.ChangeBorder();
begin
  if Assigned(FChangeBorder) then
  begin
    FChangeBorder(Self);
  end;
end;


{ TFlatPanel }

constructor TFlatPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height := 300;
  Width := 223;
  FBorders := TFlatBorders.Create();
  FCloseButton := TFlatBitBtn.Create(Self);
  FCloseButton.Parent := Self;
  FCloseButton.DrawFocus := False;
  FCloseButton.BorderStyle := FlatUtils.fbsExtraFlat;
  FCloseButton.Kind := bkCloseEx;
  FCloseButton.OnClick := DoCloseButtonClick;
  FBorderStyle := fbsFrame;
  FTopButtonOffset := 3;
  FRightButtonOffset := 3;
  RectButton;
  FCloseButton.Left := Width - FCloseButton.Width {FButtonWidth} - FRightButtonOffset;
  FCloseButton.Top := Height - (Height - FTopButtonOffset);
  //FCloseButton.Width := FButtonWidth;
  //FCloseButton.Height := FButtonHeight;
  FBorders.OnChangeBorder := RepaintBorders;
  FSeparatorHeight := 22;
  FExtraFlatCloseButton := True;
  FShowCloseButton := True;
  FShowSeparator := True;
  FEnabledCloseButton := True;
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    {csSetCaption,} csOpaque, csDoubleClicks, csReplicatable];
  //FCloseButton.TabStop := True;
  //FCloseButton.TabOrder := 0;
  TabStop := True;
end;

destructor TFlatPanel.Destroy;
begin
  FBorders.Free;
  FCloseButton.Free;
  inherited Destroy;
end;

procedure TFlatPanel.Paint;
begin
  Canvas.Brush.Color := Color;
  Canvas.FillRect(GetClientRect);
  if FShowSeparator then
  begin
    Canvas.Pen.Color := clBtnHighlight;
    Canvas.MoveTo(1, FSeparatorHeight + 1);
    Canvas.LineTo(Width - 1, FSeparatorHeight + 1);
    Canvas.Pen.Color := clBtnShadow;
    Canvas.MoveTo(1, FSeparatorHeight);
    Canvas.LineTo(Width - 1, FSeparatorHeight);
  end;
  if (FBorderStyle = fbsFlat) then
  begin
    if (FBorders.LeftBorder) then
    begin
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.MoveTo(0, 0);
      Canvas.LineTo(0, Height);
    end;
    if (FBorders.BottomBorder) then
    begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.MoveTo(0, Height - 1);
      Canvas.LineTo(Width, Height - 1);
    end;
    if (FBorders.TopBorder) then
    begin
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.MoveTo(0, 0);
      Canvas.LineTo(Width, 0);
    end;
    if (FBorders.RightBorder) then
    begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.MoveTo(Width - 1, 0);
      Canvas.LineTo(Width - 1, Height);
    end;
  end else
  begin
    if (FBorders.LeftBorder) then
    begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.MoveTo(0, 0);
      Canvas.LineTo(0, Height);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.MoveTo(1, 0);
      Canvas.LineTo(1, Height);
    end;
    if (FBorders.BottomBorder) then
    begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.MoveTo(0, Height - 2);
      Canvas.LineTo(Width, Height - 2);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.MoveTo(0, Height - 1);
      Canvas.LineTo(Width, Height - 1);
    end;
    if (FBorders.TopBorder) then
    begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.MoveTo(0, 0);
      Canvas.LineTo(Width, 0);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.MoveTo(0 + Byte(Borders.LeftBorder), 1);
      Canvas.LineTo(Width, 1);
    end;
    if (FBorders.RightBorder) then
    begin
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.MoveTo(Width - 1, 0);
      Canvas.LineTo(Width - 1, Height);
      Canvas.Pen.Color := clBtnShadow;
      Canvas.MoveTo(Width - 2, 0 + Byte(Borders.TopBorder));
      Canvas.LineTo(Width - 2, Height - 2*Byte(Borders.BottomBorder));
    end;
  end;
end;

procedure TFlatPanel.RepaintBorders(Sender: TObject);
begin
  Invalidate;
end;

procedure TFlatPanel.SetBorders(const Value: TFlatBorders);
begin
  if (FBorders <> Value) then
  begin
    FBorders := Value;
    Invalidate;
  end;
end;

procedure TFlatPanel.SetBorderStyle(const Value: TFlatBorderStyle);
begin
  if (FBorderStyle <> Value) then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TFlatPanel.SetShowSeparator(const Value: Boolean);
begin
  if (FShowSeparator <> Value) then
  begin
    FShowSeparator := Value;
    Invalidate;
  end;
end;

procedure TFlatPanel.SetSeparatorHeight(const Value: Integer);
begin
  if (FSeparatorHeight <> Value) then
  begin
    FSeparatorHeight := Value;
    Invalidate;
  end;
end;

// CloseButton

procedure TFlatPanel.SetExtraFlatCloseButton(Value: Boolean);
begin
  if FExtraFlatCloseButton <> Value then
  begin
    FExtraFlatCloseButton := Value;
    if FExtraFlatCloseButton then
    begin
      FCloseButton.BorderStyle := FlatUtils.fbsExtraFlat;
    end else
    begin
      FCloseButton.BorderStyle := FlatUtils.fbsFlat;
    end;
    Invalidate;
  end;
end;

procedure TFlatPanel.SetShowCloseButton(Value: Boolean);
begin
  if FShowCloseButton <> Value then
  begin
    FShowCloseButton := Value;
    FCloseButton.Visible := FShowCloseButton;
    Invalidate;
  end;
end;

procedure TFlatPanel.SetEnabledCloseButton(Value: Boolean);
begin
  if (FEnabledCloseButton <> Value) then
  begin
    FEnabledCloseButton := Value;
    FCloseButton.Enabled := FEnabledCloseButton;
    Invalidate;
  end;
end;

procedure TFlatPanel.DoCloseButtonClick(Sender: TObject);
begin
  if FEnabledCloseButton then
  begin
    if Assigned(FOnCloseButtonClick) then FOnCloseButtonClick(Self);
  end;
end;

procedure TFlatPanel.WMSize(var Message: TWMSize);
begin
  FCloseButton.Left := Width - FCloseButton.Width {FButtonWidth} - FRightButtonOffset;
  FCloseButton.Top := Height - (Height - FTopButtonOffset);
  Invalidate;
end;

procedure TFlatPanel.SetRightButtonOffset(Value: Integer);
begin
  if (FRightButtonOffset <> Value) then
  begin
    FRightButtonOffset := Value;
    FCloseButton.Left := Width - FCloseButton.Width{FButtonWidth} - FRightButtonOffset;
    Invalidate;
  end;
end;

procedure TFlatPanel.SetTopButtonOffset(Value: Integer);
begin
  if (FTopButtonOffset <> Value) then
  begin
    FTopButtonOffset := Value;
    FCloseButton.Top := Height - (Height - FTopButtonOffset);
    Invalidate;
  end;
end;

function TFlatPanel.RectButton: TRect;
var
  NonClientMetrics: TNonClientMetrics;
begin
  NonClientMetrics.cbSize := SizeOf(NonClientMetrics);
  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0);
  //FButtonWidth := NonClientMetrics.iCaptionWidth - 2;
  //FButtonHeight := NonClientMetrics.iCaptionHeight - 2 * 2;
  Result := Bounds(Width - FCloseButton.Width{FButtonWidth} - FRightButtonOffset, FTopButtonOffset, FCloseButton.Width{FButtonWidth}, FCloseButton.Height{FButtonHeight});
end;

{procedure TFlatPanel.DrawCloseBitmap;
var
  R: TRect;
  NonClientMetrics: TNonClientMetrics;
  X, Y: Integer;
  Bmp: TBitmap;
begin
  NonClientMetrics.cbSize := SizeOf(NonClientMetrics);
  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0);
  X := NonClientMetrics.iCaptionWidth - 2;
  Y := NonClientMetrics.iCaptionHeight - 2 * 2;
  Bmp := TBitmap.Create();
  Bmp.Width := X;
  Bmp.Height := Y;
  R := Rect(0, 0, X, Y);
  DrawFrameControl(Bmp.Canvas.Handle, R, DFC_CAPTION, DFCS_CAPTIONCLOSE or DFCS_FLAT);
  FrameRect(Bmp.Canvas.Handle, R, GetSysColorBrush(COLOR_BTNFACE));
  FCloseButton.Glyph.Assign(Bmp);
  Bmp.Free();
end;}

{function TFlatPanel.MouseOnButton: Boolean;
begin
  Result := PtInRect(RectButton, ScreenToClient(Mouse.CursorPos));
end;}

(*procedure TFlatPanel.DrawCloseButton(APressed: Boolean);
const
  Enables: array[Boolean] of Integer = (DFCS_INACTIVE, 0);
  Borders: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
var
  AFlat, ClipRgnExists: Boolean;
  R: TRect;
  DC: HDC;
  Rgn1, Rgn2: HRGN;
  BtnFaceBrush: HBRUSH;
begin
  AFlat := FExtraFlatCloseButton;
  R := RectButton;
  DC := GetWindowDC(Handle);
  BtnFaceBrush := CreateSolidBrush(GetSysColor(Integer(Color) and $7FFFFFFF));
  FillRect(DC, R, BtnFaceBrush);
  Rgn1 := 0;
  ClipRgnExists := False;
  if AFlat then
  begin
    Rgn1 := CreateRectRgn(0, 0, 0, 0);
    ClipRgnExists := GetClipRgn(DC, Rgn1) = 1;
    InflateRect(R, -1, -1);
    Rgn2 := CreateRectRgnIndirect(R);
    SelectClipRgn(DC, Rgn2);
    DeleteObject(Rgn2);
    InflateRect(R, 1, 1);
  end;

  DrawFrameControl(DC, R, DFC_CAPTION, Byte(APressed) * DFCS_PUSHED or Byte(AFlat) * DFCS_FLAT or Enables[FEnabledCloseButton]);

  if AFlat then
  begin
    if ClipRgnExists then
    begin
      SelectClipRgn(DC, Rgn1)
    end else
    begin
      SelectClipRgn(DC, 0);
    end;
    DeleteObject(Rgn1);
    if MouseOnButton then
    begin
      DrawEdge(DC, R, Borders[APressed], BF_RECT);
    end;
  end;
  DeleteObject(BtnFaceBrush);
  ReleaseDC(Handle, DC);
end;*)

procedure TFlatPanel.CMFontChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TFlatPanel.SetCloseButtonTabStop(const Value: Boolean);
begin
  if (FCloseButton.TabStop <> Value) then
  begin
    FCloseButton.TabStop := Value;
  end;
end;

function TFlatPanel.GetCloseButtonTabStop: Boolean;
begin
  Result := FCloseButton.TabStop;
end;

end.
