unit FlatSearchPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  FlatPanel, FlatBitBtn, FlatUtils, ComCtrls, StdCtrls;

type
  TFlatSearchPanel = class(TFlatPanel)
  private
    FTitle: String;
    FFocused: Boolean;
    FSearchButton: TFlatBitBtn;
    FHelpButton: TFlatBitBtn;
    FOnResetSearchBtnClick: TNotifyEvent;
    FOnHelpBtnClick: TNotifyEvent;
    FAnimate: TAnimate;
    FSearchTitle: String;
    FToolBarSeparatorLeft: Integer;
    FSearchBtn: TButton;
    FStopSearchBtn: TButton;
    function GetHelpBtnGlyph: TBitmap;
    function GetHelpBtnHint: String;
    function GetResetSearchBtnCaption: String;
    function GetResetSearchBtnGlyph: TBitmap;
    function GetResetSearchBtnHint: String;
    function GetResetSearchBtnWidth: Integer;
    function WindowRect: TRect;
    procedure SetResetSearchBtnHint(const Value: String);
    procedure SetResetSearchBtnCaption(const Value: String);
    procedure SetResetSearchBtnWidth(const Value: Integer);
    procedure SetResetSearchBtnGlyph(const Value: TBitmap);
    procedure SetHelpBtnGlyph(const Value: TBitmap);
    procedure SetHelpBtnWidth(const Value: Integer);
    procedure SetHelpBtnHint(const Value: String);
    procedure SetTitle(const Value: String);
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure SetSearchTitle(const Value: String);
    procedure SearchButtonClick(Sender: TObject); dynamic;
    procedure HelpButtonClick(Sender: TObject); dynamic;
    function GetHelpBtnWidth: Integer;
    procedure SetToolBarSeparatorLeft(const Value: Integer);
    procedure SetHelpButtonTabStop(const Value: Boolean);
    procedure SetSearchButtonTabStop(const Value: Boolean);
    function GetHelpButtonTabStop: Boolean;
    function GetSearchButtonTabStop: Boolean;
    procedure SetSearchBtn(const Value: TButton);
    procedure SetStopSearchBtn(const Value: TButton);
  protected
    procedure CreateWnd; override;
    procedure Paint; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure StartSearch;
    procedure StopSearch;
    procedure ShowUnderlining(ANeedShow: Boolean);
  published
    property Title: String read FTitle write SetTitle;
    property SearchTitle: String read FSearchTitle write SetSearchTitle;
    property ResetSearchBtnCaption: String read GetResetSearchBtnCaption write SetResetSearchBtnCaption;
    property HelpBtnHint: String read GetHelpBtnHint write SetHelpBtnHint;
    property ResetSearchBtnHint: String read GetResetSearchBtnHint write SetResetSearchBtnHint;
    property HelpBtnGlyph: TBitmap read GetHelpBtnGlyph write SetHelpBtnGlyph;
    property HelpBtnWidth: Integer read GetHelpBtnWidth write SetHelpBtnWidth;
    property ResetSearchBtnGlyph: TBitmap read GetResetSearchBtnGlyph write SetResetSearchBtnGlyph;
    property ResetSearchBtnWidth: Integer read GetResetSearchBtnWidth write SetResetSearchBtnWidth;
    property ToolBarSeparatorLeft: Integer read FToolBarSeparatorLeft write SetToolBarSeparatorLeft;
    property HelpButtonTabStop: Boolean read GetHelpButtonTabStop write SetHelpButtonTabStop;
    property SearchButtonTabStop: Boolean read GetSearchButtonTabStop write SetSearchButtonTabStop;
    property OnHelpBtnClick: TNotifyEvent read FOnHelpBtnClick write FOnHelpBtnClick;
    property OnResetSearchBtnClick: TNotifyEvent read FOnResetSearchBtnClick write FOnResetSearchBtnClick;
    property ShowHint;
    property SearchBtn: TButton read FSearchBtn write SetSearchBtn;
    property StopSearchBtn: TButton read FStopSearchBtn write SetStopSearchBtn;
  end;

procedure Register;

implementation

{$R *.RES}

procedure Register;
begin
  RegisterComponents('Market', [TFlatSearchPanel]);
end;

{ TFlatSearchPanel }

constructor TFlatSearchPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height := 300;
  Width := 235;
  FFocused := False;
  BorderStyle := FlatPanel.fbsFrame;
  SeparatorHeight := 22;
  ShowSeparator := True;
  FTitle := '';
  FSearchTitle := '   ';
  FToolBarSeparatorLeft := 70;
  // Searsh button
  FSearchButton := TFlatBitBtn.Create(Self);
  FSearchButton.Parent := Self;
  //FSearchButton.TabStop := True;
  //FSearchButton.TabOrder := 1;
  FSearchButton.Left := 3;
  FSearchButton.Top := 25;
  FSearchButton.Height := 22;
  FSearchButton.Width := 64;//54;
  FSearchButton.Margin := 5;
  FSearchButton.Hint := '';
  FSearchButton.BorderStyle := fbsExtraFlat;
  FSearchButton.Caption := FSearchButton.Hint;
  FSearchButton.OnClick := SearchButtonClick;
  FSearchButton.Glyph.LoadFromResourceName(hInstance, 'Find_button_glyph');
  // Help button
  FHelpButton := TFlatBitBtn.Create(Self);
  FHelpButton.Parent := Self;
  //FHelpButton.TabStop := True;
  //FSearchButton.TabOrder := 2;
  FHelpButton.Left := 75;//65;
  FHelpButton.Top := 25;
  FHelpButton.Height := 22;
  FHelpButton.Width := 26;
  FHelpButton.Hint := '';
  FHelpButton.BorderStyle := fbsExtraFlat;
  FHelpButton.OnClick := HelpButtonClick;
  FHelpButton.Glyph.LoadFromResourceName(hInstance, 'Help_button_glyph');
end;

procedure TFlatSearchPanel.CreateWnd();
begin
  inherited;
  // Animate
  FAnimate := TAnimate.Create(Self);
  FAnimate.Parent := Self;
  FAnimate.TabStop := False;
  if (not (csDesigning in ComponentState)) then
  begin
    FAnimate.ResName := 'FIND_AVI';
  end;
  FAnimate.Left := 13;
  FAnimate.Height := 16;
  FAnimate.Width := 16;
  FAnimate.Top := 55;
  FAnimate.Color := clWindow;
end;

function TFlatSearchPanel.WindowRect(): TRect;
begin
  Result := Rect(2, 50, Width - 2, Height - 2);
end;

procedure TFlatSearchPanel.Paint;
begin
  inherited;
  // Paint window area
  begin
    Canvas.Brush.Color := clWindow;
    Canvas.FillRect(WindowRect);
  end;
  // Title text
  Canvas.Font.Assign(Font);
  SetBkColor(Canvas.Handle, ColorToRGB(clBtnFace));
  Canvas.TextOut(8, 5, FTitle);
  // Search title
  begin
    SetBkColor(Canvas.Handle, ColorToRGB(clWindow));
    Canvas.Font.Style := Canvas.Font.Style + [fsBold];
    Canvas.TextOut(39, 57, FSearchTitle);
  end;
  // Toolbar
  begin
    Canvas.Pen.Color := clBtnHighlight;
    Canvas.MoveTo(2, 49);
    Canvas.LineTo(Width - 2, 49);
    Canvas.Pen.Color := clBtnShadow;
    Canvas.MoveTo(1, 48);
    Canvas.LineTo(Width - 1, 48);
  end;
  // Toolbar separator;
  begin
    Canvas.Pen.Color := clBtnShadow;
    Canvas.MoveTo(FToolBarSeparatorLeft, 27);
    Canvas.LineTo(FToolBarSeparatorLeft, 45);
    Canvas.Pen.Color := clBtnHighlight;
    Canvas.MoveTo(FToolBarSeparatorLeft + 1, 27);
    Canvas.LineTo(FToolBarSeparatorLeft + 1, 45);
  end;
  // Paint window separator
  begin
    Canvas.Pen.Color := clBtnFace;
    Canvas.MoveTo(13, 75);
    Canvas.LineTo(Width - 13, 75);
    Canvas.MoveTo(13, 76);
    Canvas.LineTo(Width - 13, 76);
  end;
  // Set default font for canvas
  Canvas.Font.Assign(Font);
end;

procedure TFlatSearchPanel.SearchButtonClick(Sender: TObject);
begin
  if Assigned(FOnResetSearchBtnClick) then
  begin
    FOnResetSearchBtnClick(Self);
  end;
end;

procedure TFlatSearchPanel.HelpButtonClick(Sender: TObject);
begin
  if Assigned(FOnResetSearchBtnClick) then
  begin
    FOnHelpBtnClick(Self);
  end;
end;

procedure TFlatSearchPanel.SetResetSearchBtnCaption(const Value: String);
begin
  if (FSearchButton.Caption <> Value) then
  begin
    FSearchButton.Caption := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetResetSearchBtnGlyph(const Value: TBitmap);
begin
  if (FSearchButton.Glyph <> Value) then
  begin
    FSearchButton.Glyph.Assign(Value);
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetHelpBtnGlyph(const Value: TBitmap);
begin
  if (FHelpButton.Glyph <> Value) then
  begin
    FHelpButton.Glyph.Assign(Value);
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetResetSearchBtnHint(const Value: String);
begin
  if (FSearchButton.Hint <> Value) then
  begin
    FSearchButton.Hint := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetHelpBtnHint(const Value: String);
begin
  if (FHelpButton.Hint <> Value) then
  begin
    FHelpButton.Hint := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetTitle(const Value: String);
begin
  if (FTitle <> Value) then
  begin
    FTitle := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.CMFontChanged(var Message: TMessage);
begin
  FSearchButton.Font.Assign(Font);
  FHelpButton.Font.Assign(Font);
  Invalidate;
end;

procedure TFlatSearchPanel.StartSearch;
begin
  FAnimate.Active := True;
end;

procedure TFlatSearchPanel.StopSearch;
begin
  FAnimate.Active := False;
end;

destructor TFlatSearchPanel.Destroy;
begin
  FSearchButton.Free;
  FHelpButton.Free;
  FAnimate.Free;
  inherited Destroy;
end;

procedure TFlatSearchPanel.SetSearchTitle(const Value: String);
begin
  if (FSearchTitle <> Value) then
  begin
    FSearchTitle := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.ShowUnderlining(ANeedShow: Boolean);
var
  Str: String;
begin
  if ANeedShow then
  begin
    FSearchButton.Caption := '&' + FSearchButton.Caption;
  end else
  begin
    Str := FSearchButton.Caption;
    if Copy(Str, 1, 1) = '&' then
    begin
      Delete(Str, 1, 1);
      FSearchButton.Caption := Str;
    end;
  end;
end;

procedure TFlatSearchPanel.SetResetSearchBtnWidth(const Value: Integer);
begin
  if (FSearchButton.Width <> Value) then
  begin
    FSearchButton.Width := Value;
    Invalidate;
  end;
end;

function TFlatSearchPanel.GetHelpBtnGlyph: TBitmap;
begin
  Result := FHelpButton.Glyph;
end;

function TFlatSearchPanel.GetHelpBtnHint: String;
begin
  Result := FHelpButton.Caption;
end;

function TFlatSearchPanel.GetResetSearchBtnCaption: String;
begin
  Result := FSearchButton.Caption;
end;

function TFlatSearchPanel.GetResetSearchBtnGlyph: TBitmap;
begin
  Result := FSearchButton.Glyph;
end;

function TFlatSearchPanel.GetResetSearchBtnHint: String;
begin
  Result := FSearchButton.Hint;
end;

function TFlatSearchPanel.GetResetSearchBtnWidth: Integer;
begin
  Result := FSearchButton.Width;
end;

function TFlatSearchPanel.GetHelpBtnWidth: Integer;
begin
  Result := FHelpButton.Width;
end;

procedure TFlatSearchPanel.SetHelpBtnWidth(const Value: Integer);
begin
  if (FHelpButton.Width <> Value) then
  begin
    FHelpButton.Width := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetToolBarSeparatorLeft(const Value: Integer);
begin
  if (FToolBarSeparatorLeft <> Value) then
  begin
    FToolBarSeparatorLeft := Value;
    Invalidate;
  end;
end;

procedure TFlatSearchPanel.SetHelpButtonTabStop(const Value: Boolean);
begin
  if (FHelpButton.TabStop <> Value) then
  begin
    FHelpButton.TabStop := Value;
  end;
end;

procedure TFlatSearchPanel.SetSearchButtonTabStop(const Value: Boolean);
begin
  if (FSearchButton.TabStop <> Value) then
  begin
    FSearchButton.TabStop := Value;
  end;
end;

function TFlatSearchPanel.GetHelpButtonTabStop: Boolean;
begin
  Result := FHelpButton.TabStop;
end;

function TFlatSearchPanel.GetSearchButtonTabStop: Boolean;
begin
  Result := FSearchButton.TabStop;
end;

procedure TFlatSearchPanel.SetSearchBtn(const Value: TButton);
begin
  if (FSearchBtn <> Value) then
  begin
    FSearchBtn := Value;
    if (FSearchBtn <> nil) then
    begin
      FSearchBtn.Top := 239;
      FSearchBtn.Left := 13;
      FSearchBtn.Height := 23;
      FSearchBtn.Width := 75;
    end;
  end;
end;

procedure TFlatSearchPanel.SetStopSearchBtn(const Value: TButton);
begin
  if (FStopSearchBtn <> Value) then
  begin
    FStopSearchBtn := Value;
    if (FStopSearchBtn <> nil) then
    begin
      FStopSearchBtn.Top := 239;
      FStopSearchBtn.Left := 106;
      FStopSearchBtn.Height := 23;
      FStopSearchBtn.Width := 75;
    end;
  end;
end;

procedure TFlatSearchPanel.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if ((Operation = opRemove) and (AComponent = FSearchBtn)) then
  begin
    FSearchBtn := nil;
  end;
  if ((Operation = opRemove) and (AComponent = FStopSearchBtn)) then
  begin
    FStopSearchBtn := nil;
  end;
end;

end.
